// =====================================================================
//  main.c
// =====================================================================

//  Author:         (c) 2013 Camil Demetrescu
//  License:        See the end of this file for license information
//  Created:        July 7, 2013

//  Last changed:   $Date: 2013/07/07 15:00:07 $
//  Changed by:     $Author: demetres $
//  Revision:       $Revision: 1.00 $

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/times.h>
#include <unistd.h>
#include <assert.h>
#include "filter.h"
#include "pgm.h"

// memcpy
#ifndef USEMEMCPY
#define USEMEMCPY 0
#endif

// instrumented version
#if VERSION == 3
#include "kstream.h"
kstream* ks;
#define NUM_IDS 3
#ifndef CHOSEN_K
#define CHOSEN_K 10
#endif
#endif

// filter settings
#define FACTOR (1/25.0)
#define BIAS (0)
FT blur_filter[25] = {
    1, 1, 1, 1, 1,
    1, 1, 1, 1, 1,
    1, 1, 1, 1, 1,
    1, 1, 1, 1, 1,
    1, 1, 1, 1, 1,
};

// time decl
typedef struct {
  double time_scale;
  clock_t tick_start, tick_stop;
  struct tms time_start, time_stop;    
} time_struct_t;

static void _start_clock(time_struct_t* t);
static void _stop_clock(time_struct_t* t);


// ---------------------------------------------------------------------
// main
// ---------------------------------------------------------------------
int main(int argc, char** argv) {

    time_struct_t t;
    unsigned char *img_in = NULL, *mask = NULL, *buffer = NULL;
    int in_rows, in_cols, mask_rows, mask_cols, num_iter, err = 0, i;
    char* out_file = NULL;

  #if VERSION == 3
    char* log_file = argv[4];
  #endif

    // check input parameters
  #if VERSION == 3
    if (argc < 5 || argc > 6) {
        fprintf(stderr,"Syntax: %s in-file mask-file "
                       "num-iter log-file [out-file]\n", argv[0]);
        return 1;
    }
  #else
    if (argc < 4 || argc > 5) {
        fprintf(stderr,"Syntax: %s in-file mask-file "
                       "num-iter [out-file]\n", argv[0]);
        return 1;
    }
  #endif

  #if VERSION == 3
    log_file = argv[4];
    if (argc == 6) out_file = argv[5];
  ks = kstream_new("test", NUM_IDS, CHOSEN_K);
  assert(ks != NULL);
  #else
    if (argc == 5) out_file = argv[4];
  #endif

    // load input image
    if (pgm_load(&img_in, &in_rows, &in_cols, argv[1])){
        fprintf(stderr,"*** Error loading input image %s\n", argv[1]);
        err = 1;
        goto cleanup;
    }
    
    // load mask
    if (pgm_load(&mask, &mask_rows, &mask_cols, argv[2])) {
        fprintf(stderr,"*** Error loading mask %s\n", argv[2]);
        err = 1;
        goto cleanup;
    }

    // check input image and mask
    if (in_rows != mask_rows || in_cols != mask_cols) {
        fprintf(stderr,"*** Image and mask size do not match\n");
        err = 1;
        goto cleanup;
    }

    // allocate image buffer
    buffer = malloc(in_rows*in_cols*sizeof(unsigned char));
    if (buffer == NULL) {
        fprintf(stderr,"*** Cannot allocate output image\n");
        err = 1;
        goto cleanup;        
    }
    
    // trim mask
    trim_mask(mask, in_rows, in_cols);
    
    // get number of iterations
    num_iter = atoi(argv[3]);

    //memcpy(buffer, img_in, in_rows*in_cols);

    // repeatedly apply filter
    _start_clock(&t);
    for (i = 0; i < num_iter; i++) {

        unsigned char* temp;

        // apply filter
        filter_conv(img_in, buffer, mask, blur_filter,
                    FACTOR, BIAS, in_rows, in_cols);

        // swap buffers
        temp   = img_in;
        img_in = buffer;
        buffer = temp;
    }
    _stop_clock(&t);

    #if VERSION == 3
    kstream_dump(ks, log_file);
    kstream_delete(ks);
    #endif

    // save output image, if requested
    if (out_file)
        if (pgm_save(img_in, in_rows, in_cols, out_file)) {
            fprintf(stderr,"*** Error saving output image %s\n", out_file);
            err = 1;
            goto cleanup;            
        }
    
cleanup:
    if (img_in)  free(img_in);
    if (mask)    free(mask);
    if (buffer)  free(buffer);

    return err;
}


// ---------------------------------------------------------------------
// _start_clock
// ---------------------------------------------------------------------
static void _start_clock(time_struct_t* t) {
    t->time_scale = 1.0 / sysconf(_SC_CLK_TCK);
    t->tick_start = times(&(t->time_start));
}


// ---------------------------------------------------------------------
// _stop_clock
// ---------------------------------------------------------------------
static void _stop_clock(time_struct_t* t) {
    t->tick_stop = times(&(t->time_stop));
    
    printf( "stopClock:\n"
           "\tusr: %7.2f sys: %7.2f tot: %7.2f wct: %7.2f\n\n",
           (t->time_stop.tms_utime - t->time_start.tms_utime) * t->time_scale,
           (t->time_stop.tms_stime - t->time_start.tms_stime) * t->time_scale,
           (t->time_stop.tms_utime - t->time_start.tms_utime +
            t->time_stop.tms_stime - t->time_start.tms_stime) * t->time_scale,
           (t->tick_stop           - t->tick_start          ) * t->time_scale);
}


// Copyright (C) 2013 Camil Demetrescu

// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.

// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
// USA
